#include <stdio.h>
#include <windows.h>
#pragma comment(lib, "user32")

void
initKeyEvent(INPUT_RECORD *r, BOOL kd, WORD vk, WCHAR uc, DWORD ks)
{
    r->EventType = KEY_EVENT;
    r->Event.KeyEvent.bKeyDown = kd;
    r->Event.KeyEvent.wRepeatCount = 1;
    r->Event.KeyEvent.wVirtualKeyCode = vk;
    r->Event.KeyEvent.wVirtualScanCode = vk ? MapVirtualKeyW(vk, 0) : 0;
    r->Event.KeyEvent.uChar.UnicodeChar = uc;
    r->Event.KeyEvent.dwControlKeyState = ks;
}

void
writeInput(HANDLE hStdin, wchar_t *str, size_t cch, int bPaste)
{
    DWORD i, j, n;
    if (!bPaste) {
        n = 2 * cch;
    } else {
        for (i = 0, n = 0; i < cch; i++) {
            n += (VkKeyScanW(str[i]) != -1) ? 2 : 6;
        }
    }
    INPUT_RECORD *input = calloc(sizeof(INPUT_RECORD), n);
    for (i = 0, j = 0; i < cch; i++) {
        if (!bPaste || VkKeyScanW(str[i]) != -1) {
            // normal key press
            initKeyEvent(&input[j++], TRUE, 0, str[i], 0);
            initKeyEvent(&input[j++], FALSE, 0, str[i], 0);
        } else {
            // Alt+Numpad paste sequence. For this test, just hard
            // code "?" (63) as the OEM character. It shouldn't matter.
            initKeyEvent(&input[j++], TRUE, VK_MENU, 0, LEFT_ALT_PRESSED);
            initKeyEvent(&input[j++], TRUE, VK_NUMPAD6, 0, LEFT_ALT_PRESSED);
            initKeyEvent(&input[j++], FALSE, VK_NUMPAD6, 0, LEFT_ALT_PRESSED);
            initKeyEvent(&input[j++], TRUE, VK_NUMPAD3, 0, LEFT_ALT_PRESSED);
            initKeyEvent(&input[j++], FALSE, VK_NUMPAD3, 0, LEFT_ALT_PRESSED);
            initKeyEvent(&input[j++], FALSE, VK_MENU, str[i], 0);
        }
    }
    FlushConsoleInputBuffer(hStdin);
    if (!WriteConsoleInputW(hStdin, input, n, &i)) {
        fprintf(stderr, "WriteConsoleInputW failed (%d)\n", GetLastError());
        exit(1);
    }
    free(input);
}

void
showInput(HANDLE hStdin)
{
    DWORD i, j;
    GetNumberOfConsoleInputEvents(hStdin, &i);
    INPUT_RECORD *pread_input = malloc(i * sizeof(INPUT_RECORD));
    if (!PeekConsoleInputW(hStdin, pread_input, i, &i)) {
        fprintf(stderr, "PeekConsoleInputW failed (%d)\n", GetLastError());
        exit(1);
    }
    printf("input records:\n");
    for (j=0; j < i; j++) {
        if (pread_input[j].EventType != KEY_EVENT) {
            continue;
        }
        KEY_EVENT_RECORD *e = &pread_input[j].Event.KeyEvent;
        printf("    vk: %02x, kd: %d, ks: %04x, uc: %04x\n",
            e->wVirtualKeyCode, e->bKeyDown, e->dwControlKeyState,
            e->uChar.UnicodeChar);
    }
    free(pread_input);
}

void
showStream(HANDLE hStdin, size_t cch)
{
    DWORD i, j;
    wchar_t *buf = calloc(sizeof(wchar_t), cch);
    if (!ReadConsoleW(hStdin, buf, cch, &i, NULL)) {
        printf("\nReadConsoleW failed (%d)\n", GetLastError());
        FlushConsoleInputBuffer(hStdin);
    } else {
        printf("stream (%d): L\"", i);
        for (j=0; j < i; j++) {
            if (buf[j] < 32 || buf[j] > 126) {
                printf("\\u%04x", buf[j]);
            } else {
                printf("%c", buf[j]);
            }
        }
        printf("\"\n");
    }
    free(buf);
    // Cross-check the screen-buffer when echo is enabled.
    if (!GetConsoleMode(hStdin, &i) || !(i & ENABLE_ECHO_INPUT)) {
        return;
    }
    WCHAR scr[10];
    CONSOLE_SCREEN_BUFFER_INFO csbi;
    HANDLE hStdout = GetStdHandle(STD_OUTPUT_HANDLE);
    if (GetConsoleScreenBufferInfo(hStdout, &csbi)) {
        COORD p = {0, csbi.dwCursorPosition.Y - 2};
        if (ReadConsoleOutputCharacterW(hStdout, scr, _countof(scr), p, &i)) {
            printf("screen: L\"");
            for (j=0; j < i; j++) {
                if (scr[j] < 32 || scr[j] > 126) {
                    printf("\\u%04x", scr[j]);
                } else {
                    printf("%c", scr[j]);
                }
            }
            printf("\"\n");
        }
    }
}

int
wmain(int argc, wchar_t **argv)
{
    DWORD mode, initMode;
    BOOL bVerbose = FALSE;
    HANDLE hStdin = GetStdHandle(STD_INPUT_HANDLE);

    if (!FlushConsoleInputBuffer(hStdin)) {
        switch (GetLastError()) {
        case ERROR_ACCESS_DENIED:
            fprintf(stderr,
                "stdin must have write access, e.g. do not use CON.\n");
            break;
        case ERROR_INVALID_HANDLE:
            fprintf(stderr, "stdin must be console input, e.g. CONIN$.\n");
        }
        return 1;
    }

    if (argc > 1 && !wcscmp(argv[1], L"-v")) {
        bVerbose = TRUE;
    }

    GetConsoleMode(hStdin, &initMode);
    mode = initMode | ENABLE_ECHO_INPUT;
    SetConsoleMode(hStdin, mode);

    wchar_t str[] = u"\U0001F61E"; // UTF-16 {0xD83D, 0xDE1E}
    size_t cch = _countof(str);
    str[cch - 1] = '\r'; // replace NUL with CR

    printf("\nTest normal with ECHO ON\n");
    writeInput(hStdin, str, cch, FALSE);
    if (bVerbose)
        showInput(hStdin);
    showStream(hStdin, cch + 1); // CR -> CRLF, so plus one

    printf("\nTest paste with ECHO ON\n");
    writeInput(hStdin, str, cch, TRUE);
    if (bVerbose)
        showInput(hStdin);
    showStream(hStdin, cch + 1);

    SetConsoleMode(hStdin, mode & ~ENABLE_ECHO_INPUT);

    printf("\nTest normal with ECHO OFF\n");
    writeInput(hStdin, str, cch, FALSE);
    if (bVerbose)
        showInput(hStdin);
    showStream(hStdin, cch + 1);

    printf("\nTest paste with ECHO OFF\n");
    writeInput(hStdin, str, cch, TRUE);
    if (bVerbose)
        showInput(hStdin);
    showStream(hStdin, cch + 1);

    FlushConsoleInputBuffer(hStdin);
    SetConsoleMode(hStdin, initMode);
    return 0;
}
